package oclib

import (
	"errors"
	"strings"

	"cloud.o-forge.io/core/oc-lib/dbs"
	"cloud.o-forge.io/core/oc-lib/dbs/mongo"
	"cloud.o-forge.io/core/oc-lib/models/resources"
	"cloud.o-forge.io/core/oc-lib/models/utils"
	"cloud.o-forge.io/core/oc-lib/models/workflow_execution"
	"cloud.o-forge.io/core/oc-lib/models/workspace"
	cron "github.com/robfig/cron/v3"
)

type workflowMongoAccessor struct {
	utils.AbstractAccessor
}

func New() *workflowMongoAccessor {
	return &workflowMongoAccessor{}
}

func (wfa *workflowMongoAccessor) getExecutions(id string, data *Workflow) ([]*workflow_execution.WorkflowExecution, error) {
	workflows_execution := []*workflow_execution.WorkflowExecution{}
	if data.Schedule != nil {
		if data.Schedule.Start == nil {
			return workflows_execution, errors.New("should get a start date on the scheduler.")
		}
		if data.Schedule.End != nil && data.Schedule.End.IsZero() {
			data.Schedule.End = nil
		}
		if len(data.Schedule.Cron) > 0 {
			if data.Schedule.End == nil {
				return workflows_execution, errors.New("a cron task should got a end date.")
			}
			cronStr := strings.Split(data.Schedule.Cron, " ")
			if len(cronStr) < 6 {
				return nil, errors.New("Bad cron message: " + data.Schedule.Cron + ". Should be at least ss mm hh dd MM dw")
			}
			subCron := strings.Join(cronStr[:6], " ")
			// cron should be parsed as ss mm hh dd MM dw t (min 6 fields)
			specParser := cron.NewParser(cron.Second | cron.Minute | cron.Hour | cron.Dom | cron.Month | cron.Dow)
			sched, err := specParser.Parse(subCron)
			if err != nil {
				return workflows_execution, errors.New("Bad cron message: " + err.Error())
			}
			for s := sched.Next(*data.Schedule.Start); !s.IsZero() && s.Before(*data.Schedule.End); s = sched.Next(s) {
				obj := &workflow_execution.WorkflowExecution{
					AbstractObject: utils.AbstractObject{
						Name: data.Schedule.Name,
					},
					ExecDate:   &s,
					EndDate:    data.Schedule.End,
					State:      1,
					WorkflowID: id,
				}
				workflows_execution = append(workflows_execution, obj)
			}

		} else {
			obj := &workflow_execution.WorkflowExecution{
				AbstractObject: utils.AbstractObject{
					Name: data.Schedule.Name,
				},
				ExecDate:   data.Schedule.Start,
				EndDate:    data.Schedule.End,
				State:      1,
				WorkflowID: id,
			}
			workflows_execution = append(workflows_execution, obj)
		}
	}
	return workflows_execution, nil
}

func (wfa *workflowMongoAccessor) DeleteOne(id string) (utils.DBObject, int, error) {
	return wfa.GenericDeleteOne(id, wfa)
}

func (wfa *workflowMongoAccessor) execution(id string, realData *Workflow, delete bool) (int, error) {
	if realData.Schedule == nil {
		return 200, nil
	}
	res, _, _ := wfa.LoadOne(id)
	r := res.(*Workflow)
	if r.Schedule.Start == realData.Schedule.Start && r.Schedule.End == realData.Schedule.End && r.Schedule.Cron == realData.Schedule.Cron {
		return 200, nil
	}
	if !realData.CheckBooking() {
		return 409, errors.New("the booking is already taken.")
	}
	if delete {
		mongo.MONGOService.DeleteMultiple(map[string]interface{}{
			"workflow_id": id,
			"state":       1,
		}, utils.WORKFLOW_EXECUTION.String())
	}
	accessor := (&workflow_execution.WorkflowExecution{}).GetAccessor()
	execs, err := wfa.getExecutions(id, realData)
	if err == nil && len(execs) > 0 {
		for _, obj := range execs {
			_, code, err := accessor.StoreOne(obj)
			if code != 200 {
				return code, err
			}
		}
	} else {
		return 422, err
	}
	return 200, nil
}

func (wfa *workflowMongoAccessor) UpdateOne(set utils.DBObject, id string) (utils.DBObject, int, error) {
	if code, err := wfa.execution(id, set.(*Workflow), true); err != nil {
		return nil, code, err
	}
	res, code, err := wfa.GenericUpdateOne(set, id, wfa, &Workflow{})
	if code != 200 {
		return nil, code, err
	}
	wfa.execute(res.(*Workflow))
	return res, code, err
}

func (wfa *workflowMongoAccessor) StoreOne(data utils.DBObject) (utils.DBObject, int, error) {
	res, code, err := wfa.GenericStoreOne(data, wfa)
	if err != nil {
		return nil, code, err
	}
	if code, err := wfa.execution(res.GetID(), res.(*Workflow), false); err != nil {
		return nil, code, err
	}
	wfa.execute(res.(*Workflow))
	return res, code, err
}

func (wfa *workflowMongoAccessor) CopyOne(data utils.DBObject) (utils.DBObject, int, error) {
	return wfa.GenericStoreOne(data, wfa)
}

func (wfa *workflowMongoAccessor) execute(workflow *Workflow) {
	accessor := (&workspace.Workspace{}).GetAccessor()
	filters := &dbs.Filters{
		Or: map[string][]dbs.Filter{
			"abstractobject.name": {{dbs.LIKE.String(), workflow.Name + "_workspace"}},
		},
	}
	resource, _, err := accessor.Search(filters, "")
	if err == nil && len(resource) > 0 {
		accessor.UpdateOne(&workspace.Workspace{
			Active: true,
			ResourceSet: resources.ResourceSet{
				Datas:       workflow.Datas,
				Processings: workflow.Processings,
				Storages:    workflow.Storages,
				Workflows:   workflow.Workflows,
				Datacenters: workflow.Datacenters,
			},
		}, resource[0].GetID())
	} else {
		accessor.StoreOne(&workspace.Workspace{
			Active:         true,
			AbstractObject: utils.AbstractObject{Name: workflow.Name + "_workspace"},
			ResourceSet: resources.ResourceSet{
				Datas:       workflow.Datas,
				Processings: workflow.Processings,
				Storages:    workflow.Storages,
				Workflows:   workflow.Workflows,
				Datacenters: workflow.Datacenters,
			},
		})
	}
}

func (wfa *workflowMongoAccessor) LoadOne(id string) (utils.DBObject, int, error) {
	var workflow Workflow
	res_mongo, code, err := mongo.MONGOService.LoadOne(id, wfa.GetType())
	if err != nil {
		wfa.Logger.Error().Msg("Could not retrieve " + id + " from db. Error: " + err.Error())
		return nil, code, err
	}
	res_mongo.Decode(&workflow)
	wfa.execute(&workflow)
	return &workflow, 200, nil
}

func (wfa workflowMongoAccessor) LoadAll() ([]utils.ShallowDBObject, int, error) {
	objs := []utils.ShallowDBObject{}
	res_mongo, code, err := mongo.MONGOService.LoadAll(wfa.GetType())
	if err != nil {
		wfa.Logger.Error().Msg("Could not retrieve any from db. Error: " + err.Error())
		return nil, code, err
	}
	var results []Workflow
	if err = res_mongo.All(mongo.MngoCtx, &results); err != nil {
		return nil, 404, err
	}
	for _, r := range results {
		objs = append(objs, &r.AbstractObject)
	}
	return objs, 200, nil
}

func (wfa *workflowMongoAccessor) Search(filters *dbs.Filters, search string) ([]utils.ShallowDBObject, int, error) {
	objs := []utils.ShallowDBObject{}
	if (filters == nil || len(filters.And) == 0 || len(filters.Or) == 0) && search != "" {
		filters = &dbs.Filters{
			Or: map[string][]dbs.Filter{
				"abstractobject.name": {{Operator: dbs.LIKE.String(), Value: search}},
			},
		}
	}
	res_mongo, code, err := mongo.MONGOService.Search(filters, wfa.GetType())
	if err != nil {
		wfa.Logger.Error().Msg("Could not store to db. Error: " + err.Error())
		return nil, code, err
	}
	var results []Workflow
	if err = res_mongo.All(mongo.MngoCtx, &results); err != nil {
		return nil, 404, err
	}
	for _, r := range results {
		objs = append(objs, &r)
	}
	return objs, 200, nil
}
